#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
For each glacier: selecting the most favored ARMA model among all the members
Detrending is done separately pre-DateBreakTrend and post-DateBreakTrend

ARMA model selected is ARMA(p*,q*)
Fit ARMA(p*,q*) to the residual time series of each member
Find the ARMA* model among members with the max likelihood
Save the ARMA* model coefficients

If tie between the most favored ARMA models: ARMA* is chosen as the lowest order
one among the most favored

@author: vincent
"""

import os
import sys
import copy
import csv
import numpy as np

from codeFunctions import piecewisePolyFitSingleBreak
from codeFunctions import getDevFromYearlyMean
from codeFunctions import fromMonthlyToAnnual
from codeFunctions import fitAndGenerateARMA
from codeFunctions import evaluateARMAmodel

saveParamCombs   = True
cwd              = os.getcwd()+'/'

SelModel             = 'MIROCES2L'
SelExperiment        = 'hist2100ssp585'
SelCriterion         = 'BIC' #'SigParam'/'BIC'
FitCoef              = 'Averaged' #'BestMember'/'Averaged'

### Parameters ###
SsnDateBreak  = 2015
TrendNbBreaks = 1
TrendOrderFit = 2
SsnOrderFit   = 1

algorithm        = 'innovations_mle' #algorithm for ARMA tuning by statsmodel
alpha            = 0.1 #level used for evaluating significance of parameters
DepthRange       = [0,500]
ShallowThreshold = 100
QMperiodObs      = [1950,2015]

possibleCombs       = []
fullcountForSigPrm  = []
fullcountForBIC     = []
           
DirResInsh       = cwd+'InputOutput/'
fileCombsARMA    = f'{DirResInsh}'+f'combsARMAsel_allmembersallglaciersinshore_{SelModel}_{SelExperiment}.csv'
InshModTFloc     = f'{DirResInsh}inshoreTFseries{SelModel}_{SelExperiment}_QMen4anlnrnbTFdpavg_Dp{DepthRange[0]}to{DepthRange[1]}_bathymin{ShallowThreshold}_ObsPer{QMperiodObs[0]}to{QMperiodObs[1]}.csv'
infoInshModloc   = f'{DirResInsh}infoIDinshoreTFseries{SelModel}_{SelExperiment}_QMen4anlnrnbTFdpavg_Dp{DepthRange[0]}to{DepthRange[1]}_bathymin{ShallowThreshold}_ObsPer{QMperiodObs[0]}to{QMperiodObs[1]}.csv'
LocOutput        = f'{DirResInsh}/fitARMAparams_allglaciersinshore_{SelModel}_{SelExperiment}_criterion{SelCriterion}_fitcoef{FitCoef}.csv'

dataCombs        = np.genfromtxt(fileCombsARMA,dtype=str,delimiter=',',skip_header=1)
ndC              = np.shape(dataCombs)[0]
fullGlaciers     = dataCombs[:,1]
uniqueGlaciers   = np.unique(fullGlaciers)
nGl              = len(uniqueGlaciers)
fullMembers      = dataCombs[:,2]
uniqueMembers    = np.unique(fullMembers)
nMb              = len(uniqueMembers)
arordSigPrm      = dataCombs[:,3].astype(float).astype(int)
maordSigPrm      = dataCombs[:,4].astype(float).astype(int)
arordBIC         = dataCombs[:,5].astype(float).astype(int)
maordBIC         = dataCombs[:,6].astype(float).astype(int)
uniquearords     = np.unique(np.append(arordSigPrm,arordBIC))
uniquemaords     = np.unique(np.append(maordSigPrm,maordBIC))
for oo0 in uniquearords:
    for oo1 in uniquemaords:
        possibleCombs.append([oo0,oo1])
combsarray     = np.array(possibleCombs)
porderarray    = combsarray[:,0]
qorderarray    = combsarray[:,1]
# Counting for each glacier of SelModel #
countsSigParam = np.zeros((nGl,len(possibleCombs)))
countsBIC      = np.zeros((nGl,len(possibleCombs)))
for rr in range(ndC):
    ggrow                       = np.where(uniqueGlaciers==fullGlaciers[rr])[0][0]
    iSP                         = np.where(np.logical_and(combsarray[:,0]==arordSigPrm[rr],combsarray[:,1]==maordSigPrm[rr]))[0][0]
    countsSigParam[ggrow,iSP]  += 1
    iBIC                        = np.where(np.logical_and(combsarray[:,0]==arordBIC[rr],combsarray[:,1]==maordBIC[rr]))[0][0]
    countsBIC[ggrow,iBIC]      += 1

### Find the most favored ARMA model among all the members ###
CombPerGlacier = np.zeros((nGl,2)).astype(int)
if(SelCriterion=='SigParam'):
    countsForSel = np.copy(countsSigParam)
elif(SelCriterion=='BIC'):
    countsForSel = np.copy(countsBIC)
else:
    print(f'Error: {SelCriterion} not implemented')
    sys.exit()
for rr,row in enumerate(countsForSel):
    sel0 = row==max(row)
    if(np.sum(sel0)==1):
        # Unique favored combination #
        CombPerGlacier[rr,:] = combsarray[sel0,:]
    elif(np.sum(sel0)>1):
        # Tie: find lowest order among the leaders #
        minparamnumber = np.min(np.sum(combsarray[sel0,:],axis=1))
        sel1 = np.logical_and(sel0,np.sum(combsarray,axis=1)==minparamnumber)
        if(np.sum(sel1)==1):
            # Favored combination found #
            CombPerGlacier[rr,:] = combsarray[sel1,:]
        elif(np.sum(sel1)>1):
            # Several with same number of params: select lowest AR order #
            minarorder = np.min(combsarray[sel1,0])
            sel2 = np.logical_and(sel1,combsarray[:,0]==minarorder)
            CombPerGlacier[rr,:] = combsarray[sel2,:]

### Load TF time series and Info ###
dataTFinshMod    = np.genfromtxt(InshModTFloc,dtype=float,delimiter=',')
timing           = dataTFinshMod[0,:]
stepsperyr       = 12
srTFinshMod      = dataTFinshMod[1:,:]
datainfoinshMod  = np.genfromtxt(infoInshModloc,dtype=str,delimiter=',')
modelInshore     = datainfoinshMod[:,0]
memberInshore    = datainfoinshMod[:,1]
glacierInshore   = datainfoinshMod[:,2]

### Fit coefficients for each glacier time series ###
ls_Glaciers               = []
ls_ARcoeffsGlaciers       = []
ls_MAcoeffsGlaciers       = []
ls_Sig2coeffsGlaciers     = []
ls_IntcptPreGlaciers      = []
ls_IntcptPostGlaciers     = []
ls_TrendLinPreGlaciers    = []
ls_TrendLinPostGlaciers   = []
ls_TrendQuadPreGlaciers   = []
ls_TrendQuadPostGlaciers  = []
ls_TrendBrkGlaciers       = []
ls_SsnIntcptPreGlaciers   = []
ls_SsnIntcptPostGlaciers  = []
ls_SsnTrendPreGlaciers    = []
ls_SsnTrendPostGlaciers   = []
ls_SsnBrkGlaciers         = []
ls_MeanGlaciers           = []
ls_MemberChoice           = []
### Loop over each glacier ###
for gg,glac in enumerate(uniqueGlaciers):
    ls_Glaciers.append(glac)
    # Selected AR and MA orders #
    selectedARorder    = CombPerGlacier[gg,0]
    selectedMAorder    = CombPerGlacier[gg,1]
    print(f'Calibrating an ARMA({selectedARorder},{selectedMAorder}), Glacier number {gg}')
    ggbicvals    = np.zeros(nMb)
    ggarparams   = np.zeros((nMb,max([1,selectedARorder])))
    ggmaparams   = np.zeros((nMb,max([1,selectedMAorder])))
    ggsig2param  = np.zeros(nMb)
    ### Loop over each member ###
    for mm,member in enumerate(uniqueMembers):
        # Extract inshore time series #
        ind     = np.intersect1d(np.where(memberInshore==member)[0],np.where(glacierInshore==glac)[0])[0]
        srorg   = srTFinshMod[ind,:]
        mnsrorg = np.mean(srorg)
        
        ### Remove trend with breakpoint ###
        if(TrendNbBreaks==1 and TrendOrderFit==2):
            params,DateBreak,preBrk,postBrk = piecewisePolyFitSingleBreak(timing,srorg,Order=2)
            b2pre,b1pre,b0pre,b2post,b1post,b0post = params
        srTrMn   = np.append(b0pre+b1pre*(timing[preBrk]-timing[0])+b2pre*(timing[preBrk]-timing[0])**2,
                                 b0post+b1post*(timing[postBrk]-DateBreak)+b2post*(timing[postBrk]-DateBreak)**2)
        srdtr    = mnsrorg+(srorg-srTrMn) 
        mnsrdtr  = np.mean(srdtr)

        ### Remove trended monthly effects with breakpoint ###
        fullmonthlydevs    = getDevFromYearlyMean(timing,srdtr) #monthly deviation wrt yearly average (different for each year)
        srSsn              = np.zeros(len(timing))
        if(SsnOrderFit==1):
            for mth in range(12):
                # Find all indices of month mth #
                imth          = np.zeros(12).astype(bool)
                imth[mth]     = True
                imth          = np.tile(imth,len(timing/12)+1)[0:len(timing)]
                params,brk,iipre,iipost   = piecewisePolyFitSingleBreak(timing[imth],fullmonthlydevs[imth],Order=1,BreakPoint=SsnDateBreak)
                b1pre,b0pre,b1post,b0post = params
                # Reconstruct seasonality #
                srSsn[imth]  = np.append(b0pre+b1pre*(timing[imth][iipre]-timing[imth][0]),
                                         b0post+b1post*(timing[imth][iipost]-brk))
        
        ### Compute residual time series ###
        srRsd = srorg-srTrMn-srSsn
        # Annual averaging of residuals #
        timingyr,srRsdyr = fromMonthlyToAnnual(timing,srRsd)
        
        # Calibrate the ARMA model #
        try:
            armadic      = fitAndGenerateARMA(timingyr,srRsdyr,porder=selectedARorder,qorder=selectedMAorder,nsamples=0,trendType='n',printSummary=False,method=algorithm)
            finalbic,_   = evaluateARMAmodel(timingyr,srRsdyr,alphalevel=alpha,parray=selectedARorder,qarray=selectedMAorder,trendType='n',printSummary=False,method=algorithm)
            ggbicvals[mm] = float(finalbic)
            # Retrieve the parameter estimates #
            if(selectedARorder>=1):
                ggarparams[mm,:] = armadic['ARlagparameters']
            if(selectedMAorder>=1):
                ggmaparams[mm,:] = armadic['MAlagparameters']
            ggsig2param[mm] = armadic['noisevariance']

        except:
            print(f'\nARMA fitting failed for {member} \n')
            ggbicvals[mm] = 1e20 #statsmodels ARMA fitting failed  

    ### Taking coefficients from the best fitting member ###
    if(FitCoef=='BestMember'):
        indBestMember  = np.argmin(ggbicvals)
        ls_MemberChoice.append(uniqueMembers[indBestMember])
        if(selectedARorder>=1):
            ls_ARcoeffsGlaciers.append(ggarparams[indBestMember,:])
        elif(selectedARorder==0):
            ls_ARcoeffsGlaciers.append(np.array([0]))
        if(selectedMAorder>=1):
            ls_MAcoeffsGlaciers.append(ggmaparams[indBestMember,:])
        elif(selectedMAorder==0):
            ls_MAcoeffsGlaciers.append(np.array([0]))
        ls_Sig2coeffsGlaciers.append(ggsig2param[indBestMember])
    
    ### Taking the average coefficients ###
    elif(FitCoef=='Averaged'):
        ls_MemberChoice.append('Averaged')
        indsSel = ggbicvals<1e10
        if(selectedARorder>=1):
            ls_ARcoeffsGlaciers.append(np.mean(ggarparams[indsSel,:],axis=0))
        elif(selectedARorder==0):
            ls_ARcoeffsGlaciers.append(np.array([0]))
        if(selectedMAorder>=1):
            ls_MAcoeffsGlaciers.append(np.mean(ggmaparams[indsSel,:],axis=0))
        elif(selectedMAorder==0):
            ls_MAcoeffsGlaciers.append(np.array([0]))
        ls_Sig2coeffsGlaciers.append(np.mean(ggsig2param[indsSel]))

    
    ### Get trend and monthly effects from the ensemble mean ###
    gginds          = glacierInshore==glac
    ensMn           = np.mean(srTFinshMod[gginds,:],axis=0)
    if(TrendNbBreaks==1 and TrendOrderFit==2):
        params,DateBreak,preBrk,postBrk = piecewisePolyFitSingleBreak(timing,ensMn,Order=2)
        b2pre,b1pre,b0pre,b2post,b1post,b0post = params
    preBrk          = timing<=DateBreak
    postBrk         = timing>DateBreak
        
    ls_IntcptPreGlaciers.append(b0pre)
    ls_TrendLinPreGlaciers.append(b1pre)
    ls_TrendQuadPreGlaciers.append(b2pre)
    ls_IntcptPostGlaciers.append(b0post)
    ls_TrendLinPostGlaciers.append(b1post)
    ls_TrendQuadPostGlaciers.append(b2post)
    ls_TrendBrkGlaciers.append(DateBreak)
    ls_MeanGlaciers.append(np.mean(ensMn))
    srTrMn   = np.append(b0pre+b1pre*(timing[preBrk]-timing[0])+b2pre*(timing[preBrk]-timing[0])**2,
                         b0post+b1post*(timing[postBrk]-DateBreak)+b2post*(timing[postBrk]-DateBreak)**2)
    ensMndtr = ensMn + np.mean(ensMn) - srTrMn
    
    fullmonthlydevs    = getDevFromYearlyMean(timing,ensMndtr) #monthly deviation wrt yearly average (different for each year)
    ssnpreb0           = np.zeros(12)
    ssnpreb1           = np.zeros(12)
    ssnpostb0          = np.zeros(12)
    ssnpostb1          = np.zeros(12)
    ssnbrk             = np.zeros(12)
    if(SsnOrderFit==1):
        srSsn = np.zeros(len(timing))
        for mth in range(12):
            # Find all indices of month mth #
            imth          = np.zeros(12).astype(bool)
            imth[mth]      = True
            imth          = np.tile(imth,len(timing/12)+1)[0:len(timing)]
            params,brk,iipre,iipost   = piecewisePolyFitSingleBreak(timing[imth],fullmonthlydevs[imth],Order=1,BreakPoint=SsnDateBreak)
            b1pre,b0pre,b1post,b0post = params
            # Reconstruct seasonality #
            ssnpreb1[mth]   = copy.deepcopy(b1pre) #trendPre
            ssnpostb1[mth]  = copy.deepcopy(b1post) #trendPost
            ssnpreb0[mth]   = copy.deepcopy(b0pre) #interceptPre
            ssnpostb0[mth]  = copy.deepcopy(b0post) #interceptPost
            ssnbrk[mth]     = copy.deepcopy(brk) #breakpoint
    
            srSsn[imth]  = np.append(b0pre+b1pre*(timing[imth][iipre]-timing[imth][0]),
                                     b0post+b1post*(timing[imth][iipost]-brk))

        ls_SsnIntcptPreGlaciers.append(ssnpreb0)
        ls_SsnIntcptPostGlaciers.append(ssnpostb0)
        ls_SsnTrendPreGlaciers.append(ssnpreb1)
        ls_SsnTrendPostGlaciers.append(ssnpostb1)
        ls_SsnBrkGlaciers.append(ssnbrk)
        

if(saveParamCombs):
    # Prepare output sublists of same length #
    lsForOutput   = []
    orderarmax    = np.max(CombPerGlacier[:,0])
    ordermamax    = np.max(CombPerGlacier[:,1])
    myheader      = ['Glacier','MemberChoice','EnsMnMean','EnsMnBrk']+\
                    ['EnsMnb0pre','EnsMnb1pre','EnsMnb2pre','EnsMnb0post','EnsMnb1post','EnsMnb2post']+\
                    ['EnsMnMonthBrk']+\
                    [f'EnsMnMonthb0pre_{kk}' for kk in range(12)]+\
                    [f'EnsMnMonthb0post_{kk}' for kk in range(12)]+\
                    [f'EnsMnMonthb1pre_{kk}' for kk in range(12)]+\
                    [f'EnsMnMonthb1post_{kk}' for kk in range(12)]+\
                    ['AR' for kk in range(orderarmax)]+\
                    ['MA' for kk in range(ordermamax)]+\
                    ['Sig2']
    for gg,glac in enumerate(uniqueGlaciers):
        numNAar   = orderarmax-CombPerGlacier[gg,0]
        if(CombPerGlacier[gg,0]==0):
            numNAar -= 1 #there is always a column for the zeroth coef, thus no need to add an extra column
        numNAma   = ordermamax-CombPerGlacier[gg,1]
        if(CombPerGlacier[gg,1]==0):
            numNAma -= 1 #there is always a column for the zeroth coef, thus no need to add an extra column
        lsForOutput.append([glac,ls_MemberChoice[gg],ls_MeanGlaciers[gg],ls_TrendBrkGlaciers[gg]]+\
                           [ls_IntcptPreGlaciers[gg],ls_TrendLinPreGlaciers[gg],ls_TrendQuadPreGlaciers[gg]]+\
                           [ls_IntcptPostGlaciers[gg],ls_TrendLinPostGlaciers[gg],ls_TrendQuadPostGlaciers[gg]]+\
                           [SsnDateBreak]+\
                            ls_SsnIntcptPreGlaciers[gg].tolist()+\
                            ls_SsnIntcptPostGlaciers[gg].tolist()+\
                            ls_SsnTrendPreGlaciers[gg].tolist()+\
                            ls_SsnTrendPostGlaciers[gg].tolist()+\
                            ls_ARcoeffsGlaciers[gg].tolist()+['NA' for kk in range(numNAar)]+\
                            ls_MAcoeffsGlaciers[gg].tolist()+['NA' for kk in range(numNAma)]+\
                            [ls_Sig2coeffsGlaciers[gg]])
    with open(LocOutput,'w',newline='') as outputcsv:
        writer = csv.writer(outputcsv)
        writer.writerow(myheader)
        for kk,sublist in enumerate(lsForOutput):
            writer.writerow(sublist)    
                
    
    
    
   
    

    
                

